--[[
Elements are wrappers for components that handle GUI interactions
Attributes:
-- Component: Wrapped component
-- x, y: Location
-- kind: String name of the component
-- parameter: Which component variable is set by the edit dialog
-- measure: String representing the parameter
-- units: Unit type string
]]

Element = class()

-- movements in the grid that lead to an attached component, split by [element orientation][direction]
Element.touching = { vertical = { [-1] = {{0, -2},
                                          {0, -1},
                                          {-1, -1}},
                                  [1] =  {{0, 2},
                                          {-1, 1},
                                          {0, 1}}
                    },
                     horizontal = { [-1] = {{-1, 0},
                                            {0, 1},
                                            {0, -1}},
                                   [1] =   {{1, 0},
                                            {1, -1},
                                            {1, 1}}
                  }}

-- Statics ------------------------------
function Element.drawCircle(gc, x, y, size, horizontal)
    if horizontal then
        gc:drawLine(x, y, x + size/4, y)
        gc:drawArc(x + size/4, y - size/4, size/2, size/2, 0, 360)
        gc:drawLine(x + size*(3/4), y, x + size, y)
    else
        gc:drawLine(x, y, x, y + size/4)
        gc:drawArc(x - size/4, y + size/4, size/2, size/2, 0, 360)
        gc:drawLine(x, y + size*(3/4), x, y + size)
    end
end

-- Instance ------------------------------
function Element:init(schematic, inverted, x, y)
    self.schematic = schematic
    self.x = x
    self.y = y
    self.inverted = inverted
end

function Element:getDrawingConstants(grid, camera)
    return self.x*grid.size - camera.x,
           math.floor(self.y/2)*grid.size - camera.y,
           grid.size
end

function Element:draw(gc, grid, camera)
      x, y, size = self:getDrawingConstants(grid, camera)
      self.drawComponent(gc, x, y, size, self.inverted,  self.y % 2 == 0)
end

function Element:attached(req_sign)
    local r = {[1] = {}, [-1] = {}}
    local set = self.y % 2 == 0 and Element.touching.horizontal or Element.touching.vertical
    for sign, v in pairs(set) do
        if (not req_sign or req_sign == sign) then
            for i, e in pairs(v) do
                local x, y = self.x + e[1], self.y + e[2]
                local a = self.schematic.elements[y][x]
                if a then
                    if oneWay or a:isAttached(self.x, self.y) then
                        table.insert(r[sign], a)
                    end
                end
            end
        end
    end
    return r
end

function Element:isAttached(x, y, req_sign)
      local set = self.y % 2 == 0 and Element.touching.horizontal or Element.touching.vertical
      for sign, v in pairs(set) do
          if (not req_sign or req_sign == sign) then
              for i, e in pairs(v) do
                  local x1, y1 = self.x + e[1], self.y + e[2]
                  if x1 == x and y1 == y then
                      return sign
                  end
              end
          end
      end
      return false
end

------------------
-- ElementWire  --
------------------

ElementWire = class(Element)

-- Statics ------------------------------

ElementWire.kind = "Wire"
ElementWire.tool = "wire"

function ElementWire.drawComponent(gc, x, y, size, inverted, horizontal)
    if horizontal then
        gc:fillRect(x, y, size, 1)
    else
        gc:fillRect(x, y, 1, size)
    end
end

-- Instance ------------------------------
function ElementWire:init(schematic, inverted, x, y, name, value)
    Element.init(self, schematic, inverted, x, y)
    self.voltage = false
end

function ElementWire:getDialogInfo()
    if not self.voltage then
        return {"??? V"}
    else
        return {Unit.pretty(self.voltage) .. "V"}
    end
end

function ElementWire:getDialogTitle()
    return "Wire"
end

----------------------
-- ElementResistor  --
----------------------

ElementResistor = class(Element)

-- Statics ------------------------------
ElementResistor.kind = "Resistor"
ElementResistor.tool = "resistor"
ElementResistor.measure = "Resistance"
ElementResistor.parameter = "resistance"

function ElementResistor.drawComponent(gc, x, y, size, inverted, horizontal)
    local zigzag = {0, 0, -size/5, 0, size/5, 0, -size/5, 0, size/5, 0, 0}
    if inverted then
         reverse(zigzag)
    end
    if horizontal then
        for i = 0, 9 do
            gc:drawLine(x + size*(i/10), y + zigzag[i+1], x + size*((i+1)/10), y + zigzag[i+2])
        end
    else
        for i = 0, 9 do
            gc:drawLine(x + zigzag[i+1], y + size*(i/10), x + zigzag[i+2], y + size*((i+1)/10))
        end
    end
end

-- Instance ------------------------------
function ElementResistor:init(schematic, inverted, x, y, name, value, unit)
    Element.init(self, schematic, inverted, x, y)
    self.component = Resistor(schematic.circuit, name, false, false, value, unit)
end

function ElementResistor:getDialogInfo()
    local voltage, current = "???", "???"
    if self.component.nodes[1] and self.component.nodes[-1] then
        if self.component.nodes[1].voltage and self.component.nodes[-1].voltage then
            voltage = math.abs(self.component.nodes[1].voltage - self.component.nodes[-1].voltage)
            current = Unit.pretty(voltage/self.component.resistance:value())
            voltage = Unit.pretty(voltage)
        end
    end
		return {self.component.resistance:str(),
						"-----",
						"V → " .. voltage .. "V",
						"I → " .. current .. "A"}
end

function ElementResistor:getDialogTitle()
		return "Resistor " .. self.component.name
end
--------------------
-- ElementVoltage --
--------------------

ElementVoltage = class(Element)

-- Statics ------------------------------
ElementVoltage.kind = "Voltage Source"
ElementVoltage.tool = "voltage"
ElementVoltage.measure = "Voltage"
ElementVoltage.parameter = "voltage"

function ElementVoltage.drawComponent(gc, x, y, size, inverted, horizontal)
    Element.drawCircle(gc, x, y, size, horizontal)
    -- relative to start coordinates, written for vertical
    local lines = {{size*(1/10), size*(2/5), -size*(1/10), size*(2/5)},
                  {size*(1/10), size*(3/5), -size*(1/10), size*(3/5)},
                  {0, size*((inverted and 7 or 3)/10), 0, size*(1/2)}}
    for i, line in ipairs(lines) do
         if horizontal then
             gc:drawLine(x + line[2], y + line[1], x + line[4], y + line[3])
         else
             gc:drawLine(x + line[1], y + line[2], x + line[3], y + line[4])
         end
    end
end

-- Instance ------------------------------
function ElementVoltage:init(schematic, inverted, x, y, name, value)
    Element.init(self, schematic, inverted, x, y)
    self.component = VoltageSource(schematic.circuit, name, false, false, value, unit)
end

function ElementVoltage:getDialogInfo()
		return {self.component.voltage:str()}
end

function ElementVoltage:getDialogTitle()
		return "Source " .. self.component.name
end

--------------------
-- ElementCurrent --
--------------------

ElementCurrent = class(Element)

-- Statics ------------------------------
ElementCurrent.kind = "Current Source"
ElementCurrent.tool = "current"
ElementCurrent.measure = "Current"
ElementCurrent.parameter = "current"

function ElementCurrent.drawComponent(gc, x, y, size, inverted, horizontal)
    Element.drawCircle(gc, x, y, size, horizontal)
    -- relative to start coordinates, written for vertical
    local lines = {{0, size*(1/3), 0, size*(2/3)},
                  {size/6, size*(1/2), 0, size*((inverted and 1 or 2)/3)},
                  {-size/6, size*(1/2), 0, size*((inverted and 1 or 2)/3)}}
    for i, line in ipairs(lines) do
         if horizontal then
             gc:drawLine(x + line[2], y + line[1], x + line[4], y + line[3])
         else
             gc:drawLine(x + line[1], y + line[2], x + line[3], y + line[4])
         end
    end
end

-- Instance ------------------------------
function ElementCurrent:init(schematic, inverted, x, y, name, value)
    Element.init(self, schematic, inverted, x, y)
    self.component = CurrentSource(schematic.circuit, name, false, false, value, unit)
end

function ElementCurrent:getDialogInfo()
    local voltage = "???"
    if self.component.nodes[1] and self.component.nodes[-1] then
        if self.component.nodes[1].voltage and self.component.nodes[-1].voltage then
            voltage = Unit.pretty(math.abs(self.component.nodes[1].voltage - self.component.nodes[-1].voltage))
        end
    end
		return {self.component.current:str(),
						"-----",
						"V → " .. voltage .. "V"}
end

function ElementCurrent:getDialogTitle()
		return "Source " .. self.component.name
end

-------------------
-- ElementGround --
-------------------

ElementGround = class(Element)

-- Statics ------------------------------
ElementGround.kind = "Ground"
ElementGround.tool = "ground"

function ElementGround.drawComponent(gc, x, y, size, inverted, horizontal)
    -- relative to start coordinates, written for vertical
    local lines = {{0, 0, 0, size*(1/6)},
                  {-size/3, size*(1/6), size/3, size*(1/6)},
                  {-size/6, size*(2/6), size/6, size*(2/6)},
                  {-size/9, size*(3/6), size/9, size*(3/6)}}

    for i, line in ipairs(lines) do
         if inverted then
            line = {line[1], size - line[2], line[3], size - line[4]}
         end
         if horizontal then
             gc:drawLine(x + line[2], y + line[1], x + line[4], y + line[3])
         else
             gc:drawLine(x + line[1], y + line[2], x + line[3], y + line[4])
         end
    end
end

-- Instance ------------------------------
function ElementGround:init(schematic, inverted, x, y, name, value, unit)
    Element.init(self, schematic, inverted, x, y)
end

function ElementGround:getDialogInfo()
		return {""}
end

function ElementGround:getDialogTitle()
		return "Ground"
end


function ElementGround:attached()
    return Element.attached(self, self.inverted and 1 or -1)
end

function ElementGround:isAttached(x, y)
    return Element.isAttached(self, x, y, self.inverted and 1 or -1)
end
