--[[
Schematic implements the GUI used to build circuits
Attributes:
-- elements: 2D representation of grid elementations
             rows alternate between horizontal and vertical elements
              { _  _  _  _ },
              {|| || || || },
              { _  _  _  _ }
-- panel: Reference to the info panel dialog
-- theme: Color theme:
-- selection: {x, y} location of selector
-- camera: Pixel shift of the schematic view
-- selected: Currently selected element
-- tool: Currently active tool
]]

Schematic = class(WWidget)

Schematic.defaultGrid = {width = 10, height = 10, size = 50}

Schematic.defaultTheme = {
    textColor         = {000, 000, 000},
    backgroundColor   = {255, 255, 255},
    borderColor       = {150, 150, 150},
    gridColor         = {235, 235, 235},
    unplacedColor     = {150, 150, 150},
    placedColor       = {050, 050, 255},
    selectorColor     = {200, 200, 200},
    selectedColor     = {255, 050, 050}
}

Schematic.creatorTools = {["wire"] = {element = ElementWire, name = "W", editor = false},
                          ["resistor"] = {element = ElementResistor, name = "R", editor = true},
                          ["voltage"] = {element = ElementVoltage, name = "V", editor = true},
                          ["current"] = {element = ElementCurrent, name = "C", editor = true},
                          ["ground"] = {element = ElementGround, name = "G", editor = false}
                         }

function Schematic:init(arg)
    Widget.init(self)

    -- Window Settings
    self.panel = arg.panel
    self.indicator = arg.indicator
    self.dw = 100
    self.dh = 100
    self.theme = arg.theme or Schematic.defaultTheme

    -- Grid / Circuit
    self.grid = arg.grid or self.defaultGrid
    self.grid.height = self.grid.height*2
    self.circuit = Circuit()

    -- Elements and Properties
    self.elements = {}
    self.toolShortcuts = {}
    self.camera = {x = 0, y = 0}
    self.selection = {x = 1, y = 1}
    self.selected = false
    self.tool = "resistor"
    self.inverted = true

    -- Initialize Elements
    self:clearCircuit()

    -- self:testCircuit()
end


function Schematic:clearCircuit()
    for y = 0, self.grid.height*2 do
        self.elements[y] = {}
        for x = 0, self.grid.width + (y % 2 == 0 and -1 or 0) do
            table.insert(self.elements[y], false)
        end
    end
end

function Schematic:draw(gc)

    gc:clipRect("set", self.x, self.y, self.w, self.h)

    -- Background
    gc:setColorRGB(unpack(self.theme.backgroundColor))
    gc:fillRect(self.x, self.y, self.w, self.h)
    self:drawGrid(gc)

    -- Border
    gc:setColorRGB(unpack(self.theme.borderColor))
    gc:drawRect(self.grid.size - self.camera.x, self.grid.size - self.camera.y,
                self.grid.width*self.grid.size-self.grid.size, self.grid.height*self.grid.size/2)

    -- Details
    self:drawElements(gc)
    self:drawSelector(gc)

    gc:clipRect("reset")

    self:updatePanel()

end

function Schematic:drawElements(gc)
    for y, row in pairs(self.elements) do
        for x, e in pairs(row) do
            if e then
                gc:setColorRGB(unpack(self:getColor(e.x, e.y)))
                e:draw(gc, self.grid, self.camera)
            end
        end
    end
end

function Schematic:getColor(x, y)
    if self.tool == "inspector" and self.selected then
        if x == self.selected.x and y == self.selected.y then
            return self.theme.selectedColor
        end
    end
    return self.theme.placedColor
end

function Schematic:drawGrid(gc)
    gc:setColorRGB(unpack(self.theme.gridColor))
    local s = self.grid.size
    local start = {x = self.x + ((self.w / 2) % s), y = self.y + ((self.h / 2) % s)}
    for i = 0, math.ceil(self.h / s) do
          gc:drawRect(0, i * s - (self.camera.y % s), self.w, 1)
    end
    for i = 0, math.ceil(self.w / s) do
        gc:drawRect(i * s - (self.camera.x % s), 0, 1, self.h)
    end
end

function Schematic:drawSelector(gc)
    -- Draw general selector
    gc:setColorRGB(unpack(self.theme.selectorColor))
    local x, y = self:locFromIndex(self.selection.x, self.selection.y)
    gc:drawArc(x, y, self.grid.size, self.grid.size, 0, 360)

    -- Draw Specific Tool
    gc:setColorRGB(unpack(self.theme.unplacedColor))
    local horizontal = self.selection.y % 2 == 0
    if horizontal then
        y = y + self.grid.size/2
    else
        x = x + self.grid.size/2
    end
    if Schematic.creatorTools[self.tool] then
        Schematic.creatorTools[self.tool].element.drawComponent(gc, x, y, self.grid.size, self.inverted, horizontal)
    end

    if self.tool == "loop" and self.loopStart then
        local x1, y1 = self:locFromIndex(self.selection.x, self.selection.y)
        local x2, y2 = self:locFromIndex(self.loopStart.x, self.loopStart.y)
        gc:setColorRGB(unpack(self.theme.selectedColor))
        gc:drawRect(math.min(x1, x2) + self.grid.size/2, math.min(y1, y2) + self.grid.size/2,
                    math.abs(x1-x2), math.abs(y1-y2))
    else
        self.loopStart = false
    end
end

function Schematic:locFromIndex(x, y)
    return x*self.grid.size - self.camera.x,
               (y/2)*self.grid.size + self.grid.size/2 - self.camera.y
end

function Schematic:updatePanel()
    local selected = self.selected
    self.panel.visible = selected
    if selected then
        local x, y, size = Element.getDrawingConstants(selected, self.grid, self.camera)
        if  x + size/2 > self.w/2 then
            self.panel.xx = 10
        else
            self.panel.xx = -10
        end
        self.panel.title = selected:getDialogTitle()
        self.panel.text = selected:getDialogInfo()
    end
end

function Schematic:insertComponent(component, x, y, ...)
    self.elements[y][x] = component(self, self.inverted, x, y, ...)
    return self.elements[y][x]
end

function Schematic:testCircuit()
    self:insertComponent(ElementWire, 2, 7)
    self:insertComponent(ElementVoltage, 2, 9, "V1", 10)
    self:insertComponent(ElementCurrent, 2, 6, "I1", 10)
    self:insertComponent(ElementWire, 3, 6)
    self:insertComponent(ElementResistor, 4, 6, "R1", 10)
    self:insertComponent(ElementWire, 5, 6)
    self:insertComponent(ElementResistor, 6, 7, "R2", 10)
    self:insertComponent(ElementWire, 6, 9)
    self:insertComponent(ElementWire, 2, 10)
    self:insertComponent(ElementWire, 3, 10)
    self:insertComponent(ElementWire, 4, 10)
    self:insertComponent(ElementWire, 5, 10)
end

function Schematic:zoom(zoom)
    self.grid.size = self.grid.size + zoom
    self.grid.size = math.max(math.min(self.grid.size, 60), 20)
    local x, y = self:locFromIndex(self.selection.x, self.selection.y)
    self.camera.x = self.camera.x - (self.w/2 - x)/2
    self.camera.y = self.camera.y - (self.h/2 - y)/2
    platform.window:invalidate()
end

function Schematic:moveSelection(x, y)
    local new = {x = self.selection.x + x, y = self.selection.y + y}
    self.selection.x = math.min(math.max(0.5, new.x), self.grid.width-0.5)
    self.selection.y = math.min(math.max(0, new.y), self.grid.height)
    self:selectAt(self.selection.x, self.selection.y)
    self:pan()
    platform.window:invalidate()
end

function Schematic:pan()
    local x, y = self:locFromIndex(self.selection.x, self.selection.y)
    if x > self.w - self.grid.size then
        self.camera.x = self.selection.x*self.grid.size - self.w + self.grid.size
    end
    if x < 0 then
        self.camera.x = self.selection.x*self.grid.size
    end
    if y > self.h - self.grid.size then
        self.camera.y = ((self.selection.y)/2)*self.grid.size - self.h + 1.5*self.grid.size
    end
    if y < 0 then
        self.camera.y = ((self.selection.y)/2)*self.grid.size + self.grid.size/2
    end
end

function Schematic:indexFromCoord(x, y)
    local x, y = x, y + 2
    if y % 2 ~= 0 then
        x = x + 0.5
    end
    return x, y
end

function Schematic:selectAt(x, y)
    local x, y = self:indexFromCoord(x, y)
    self.selected = false
    if self.elements[y] and self.elements[y][x] then
        self.selected = self.elements[y][x]
    end
end

function Schematic:onMouseDown()
    self.meDown = true
end

function Schematic:charIn(char)
    local arrow = {["4"] = "left",
                   ["2"] = "down",
                   ["6"] = "right",
                   ["8"] = "up",
    }
    if char == '-' then
        self:zoom(-5)
    elseif char == "+" or char == "=" then
        self:zoom(5)
    elseif char == "f" or char == "5" then
        self:invert()
    elseif arrow[char] then
        self:arrowKey(arrow[char])
    elseif char == "s" then
        self:solve()
    elseif self.toolShortcuts[char] then
        self.tool = self.toolShortcuts[char]
    end
end

function Schematic:invert()
    if self.tool == "inspector" and self.selected then
        self.selected.inverted = not self.selected.inverted
        self:edit()
    end
    self.inverted = not self.inverted
end

function Schematic:solve()
    local cb = CircuitBuilder(self.elements)
    if cb:buildCircuit() then
        self.indicator.color = {0, 255, 0}
        self.indicator.text = "✓"
    end
    platform.window:invalidate()
end

function Schematic:edit()
    self.indicator.color = {255, 0, 0}
    self.indicator.text = "≟"
end

function Schematic:arrowKey(arrow)
    local dir = {["left"] = {-0.5, 0},
                    ["down"] = {0, 1},
                    ["right"] = {0.5, 0},
                    ["up"] = {0, -1},
                   }
     if dir[arrow] then
         self:moveSelection(dir[arrow][1], dir[arrow][2])
     end
end

function Schematic:clearKey()
    self:clearCircuit()
    platform.window:invalidate()
end

function Schematic:mouseUp(x, y)
    self:enterKey()
end

function Schematic:rightMouseUp(x, y)
    self:backspaceKey()
end

function Schematic:enterKey()
    self:edit()
    if self.tool == "inspector" and self.selected then
        if self.creatorTools[self.selected.tool] and self.creatorTools[self.selected.tool].editor then
            self:launchEditor(self.selected)
        end
    elseif self.tool == "loop" then
        local x, y = self:indexFromCoord(self.selection.x, self.selection.y)
        if x % 1 ~= 0 then
            if not self.loopStart then
                self.loopStart = {x = self.selection.x, y = self.selection.y}
            else
                local x1, y1 = math.min(self.selection.x, self.loopStart.x), math.min(self.selection.y, self.loopStart.y)
                local x2, y2 = math.max(self.selection.x, self.loopStart.x), math.max(self.selection.y, self.loopStart.y)
                self:makeWireLoop(x1, y1, x2, y2)
                self.loopStart = false
            end
        end
    elseif Schematic.creatorTools[self.tool] then
        local x, y = self:indexFromCoord(self.selection.x, self.selection.y)
        if x % 1 == 0 then
            local c = self:insertComponent(Schematic.creatorTools[self.tool].element, x, y,
                                           Schematic.creatorTools[self.tool].name, 1, "")
            self.selected = c
            if Schematic.creatorTools[self.tool].editor then
                self:launchEditor(c)
            end
        end
    end
end

function Schematic:makeWireLoop(x1, y1, x2, y2)
    for x = x1, x2, .5 do
        for y = y1, y2 do
            if x == x1 or x == x2 or y == y1 or y == y2 then
                local wx, wy = self:indexFromCoord(x, y)
                if wx % 1 == 0 then
                    self:insertComponent(ElementWire, wx, wy, "w")
                end
            end
        end
    end
end

function Schematic:backspaceKey()
    local x, y = self:indexFromCoord(self.selection.x, self.selection.y)
    if x % 1 == 0 then
        self.elements[y][x] = false
        self.selected = false
    end
end

function Schematic:launchEditor(element)
    local editor = EditComponent(element, 20, 20, -40, -40)
    push_screen_direct(editor)
end

function Schematic:mouseMove(x, y)
    local cx, cy = self.camera.x, self.camera.y
    local mx, my = (x + cx - self.grid.size/4)/self.grid.size, (y + cy - self.grid.size/4)/(self.grid.size/2)
    self:moveSelection(math.floor((mx - self.selection.x)*2)/2,
                                 math.floor((my - self.selection.y - 1)))
end
