--[[
Components are devices used to connect nodes
Attributes:
-- Circuit: Circuit the component is part of
-- Name: A name for the component
-- nodes[1]: Node connected to the positive end of the component
-- nodes[-1]: Node connected to the negative end of the component
-- current: Current through the component (set when circuit is solved)
-- voltage: Voltage across the component (set when circuit is solved)
]]

Component = class()
-- Component Constructor
-- Takes the circuit it is part of, a name, and the connected nodes
function Component:init(circuit, name, posNode, negNode)
		self.circuit = circuit
		table.insert(self.circuit.components, self)
		self.name = name
		self.nodes = {}
		self.nodes[1] = posNode
		self.nodes[-1] = negNode
		if posNode and negNode then
				posNode:connect(self, 1)
				negNode:connect(self, -1)
		end
		self.current = Unit()
		self.voltage = Unit()
end

-- Get the node connected to a given side of the component
function Component:getNode(sign)
		return self.nodes[sign]
end

Resistor = class(Component)
Resistor.unit = "Ω"
	-- Resistor
	-- Takes in a resistance in Ohms
function Resistor:init(circuit, name, posNode, negNode, resistance, unit)
		Component.init(self, circuit, name, posNode, negNode)
		self.resistance = Unit(resistance, unit, self.unit)
end

-- Current is (v2-v1)/R
function Resistor:getCurrent(node)
		local c = '(' .. self.nodes[-node].name .. '-' .. self.nodes[node].name .. ')'
		c = c .. '/' .. self.resistance:value()
		return c
end

VoltageSource = class(Component)
VoltageSource.unit = "V"
-- Voltage Source
-- Takes in a voltage in Volts
function VoltageSource:init(circuit, name, posNode, negNode, voltage, unit)
		Component.init(self, circuit, name, posNode, negNode)
		self.voltage = Unit(voltage, unit, self.unit)
end

-- Voltage at given node = Voltage at opposite node + Voltage from source
function VoltageSource:getVoltage(node)
		return self.nodes[-node].name .. '=' .. self.nodes[node].name .. ' + ' .. self.voltage:value()*node
end


CurrentSource = class(Component)
CurrentSource.unit = "A"
-- Current Source
-- Takes in a current in Amperes
function CurrentSource:init(circuit, name, posNode, negNode, current, unit)
		Component.init(self, circuit, name, posNode, negNode)
		self.current = Unit(current, unit, self.unit)
end

-- Current from this device is simply its supplied current
function CurrentSource:getCurrent(node)
		return self.current:value()*node
end
