--[[
Nodes are ares of the circuit that are directly connected by wire
Attributes:
-- Circuit: Circuit the component is part of
-- Name: A name for the component
-- voltage: Voltage at the node (based on a grounded node used when solving)
-- connections: passive components and current sources connected
-- voltageConnections: voltage sources connected
NOTE: connections and voltageConnections are seperated because nodal analysis
      is used to solve these circuits.
]]

Node = class()

-- Nodes Constructor
function Node:init(circuit, name)
		self.circuit = circuit
		table.insert(self.circuit.nodes, self)
		self.connections = {}
		self.voltageConnections = {}
		self.voltage = false
		self.name = name
end

-- Connect a component to a node
-- (sign is 1 or -1, and is the side of component connected)
function Node:connect(component, sign)
		component.nodes[sign] = self
		if component.getVoltage then
				table.insert(self.voltageConnections, Connection(component, sign))
		else
				table.insert(self.connections, Connection(component, sign))
		end
end

-- Add together outgoing currents of connected nodes
-- Used in both Current Equation and Super Node Equation
function Node:addCurrents()
		local eq = ''
		for i, v in pairs(self.connections) do
				eq = eq .. v.component:getCurrent(v.sign) .. ' + '
		end
		return eq
end

-- Get the Current Equation
-- Used when all nodes are connected by a resistive or current source
function Node:getCurrentEq()
		local eq = '0 = 0 + ' .. self:addCurrents()
		eq = string.sub(eq, 0, #eq-2) .. ', '
		return eq
end

-- Get the Voltage Equation
-- Returns equations relating to nodes connected by a voltage source
function Node:getVoltageEq()
		local eq = ''
		for i, v in pairs(self.voltageConnections) do
				eq = eq .. v.component:getVoltage(v.sign) .. ', '
		end
		return eq
end

-- Get Super Node Equation
-- This links all the outgoing equations of nodes connected by a voltage source
function Node:getSuperEq(visited)
		local eq = ''
		local og = false
		if #visited == 1 then
				eq = eq .. '0 = 0 + '
				local og = true
		end
		eq = eq .. self:addCurrents()
		for i, v in pairs(self.voltageConnections) do
				if not visited[v.component:getNode(-v.sign).name]  then --not visited[v.component.name] then
					 -- visited[v.component.name] = true
					 	visited[v.component:getNode(-v.sign).name] = true
						eq = eq .. v.component:getNode(-v.sign):getSuperEq(v.component, visited)
				end
		end
		if og then
				eq = string.sub(eq, 0, #eq-2) .. ', '
		end
		return eq
end
